/*** simple LOKI (or LOKI alike) detection routines.
 *** (C) 1999 by Stealth.
 ***/
#include "icmp.h"

struct virtual_connections {
   	int replys, requests;
        unsigned int src, dst;
        struct virtual_connections *next;
};

struct virtual_connections *first = NULL;

void alarm_handler(int);

int
passive_loki(char *intf, int timeout)
{
   	int fd = 0;
        char buf[10000];
        struct icmp *icmp;
        struct ip *ip;
	unsigned int t = 0;

	puts(we);        
	printf("Try to detect active LOKI sessions. You gave me %d seconds.\n"
	       "Please be patient.\n\n", timeout);

        /* set handler to interrupt us later */
   	signal(SIGALRM, alarm_handler);

   	/* give us a socket */
        if ((fd = socket(PF_INET, SOCK_PACKET, htons(ETH_P_ALL))) < 0) {
           	perror("socket");
                exit(errno);
        }
        setpromisc(fd, intf);     
        promisc = fd;
        alarm(timeout);        
        while (1) {
           	bzero(buf, 10000);
           	if (read(fd, buf, 9999) < 0 && errno != EINTR) {
                   	perror("read");
                        exit(errno);
                }
                /* point to IP and ICMP headers in received packet */
                ip = (struct ip*)(buf + sizeof(struct ethhdr));
                icmp = (struct icmp*)(buf + sizeof(struct ethhdr) + (ip->ip_hl<<2));
                
                /* only handle ICMP packets */
                if (ip->ip_p != ICMPv4)
                   	continue;
                /* from type 0 and type 8 */
                if (icmp->icmp_type != ICMP_ECHOREPLY && icmp->icmp_type != ICMP_ECHO)
                   	continue;
		/* swap SRC/DST if REPLY, or we count the connections doubled */
		if (icmp->icmp_type == ICMP_ECHOREPLY) {
			t = ip->ip_src.s_addr;
			ip->ip_src.s_addr = ip->ip_dst.s_addr;
			ip->ip_dst.s_addr = t;
		}
                add_connection(ip->ip_src.s_addr,ip->ip_dst.s_addr, icmp->icmp_type);
        }                
}

int
add_connection(unsigned int src, unsigned int dst, int what)
{
   	struct virtual_connections *l;
        
        l = first;
        
        /* look if connection already exists */
        while (l) {
           	/* if so, count type */
           	if (l->src == src && l->dst == dst) {
                   	if (what == ICMP_ECHO) {
#ifdef DEBUG
				printf("requests++\n");
#endif
                           	l->requests++;
			} else {
#ifdef DEBUG
				printf("replys++\n");
#endif
                           	l->replys++;
			}
                   	return 0;
                }
                l = l->next;
        }
        /* Ok, here we are. Thus the connection doesn't exist */
        if ((l = (struct virtual_connections*)malloc(sizeof(struct virtual_connections))) == NULL) {
                perror("malloc");
                exit(errno);
        }
        if (what == ICMP_ECHO)
           	l->requests++;
        else
           	l->replys++;
        l->src = src;
        l->dst = dst;
        l->next = first;
#ifdef DEBUG
	printf("%d %d\n", src, dst);
#endif
	
        first = l;
        return 0;
}

void
alarm_handler(int sig)
{
	struct in_addr from, to;
	char *s1 = NULL, *s2 = NULL;
	int bad = 0;
	
   	while (first) {
#ifdef DEBUG
           	printf("%d %d\n", first->replys, first->requests);
#endif
		if ((first->replys / first->requests) > MAXREPLYS) {
			from.s_addr = first->dst;
			to.s_addr = first->src;
			s1 = resolv(from, 1);
			s2 = resolv(to, 1);
			printf("\a\n*******\n");	
			printf("Alert! Possibly LOKI session detected from %s to %s!\n", s2, s1);
			printf("(attacked host is %s)\n", s1);
			printf("*******\n\n");
			bad = 1;
		}	
                first = first->next;
        }
	if (!bad) 
		printf("No active sessions detected. Ok.\n\n");
	free(s1);
	free(s2);
        droppromisc(promisc);
        exit(0);
}

